/*
 * tsb.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

/*
 * USAGE:	% tsb [-xXCDOvV] [-S LEV] -f filename [-I inputlistfile | filemask ...]
 *
 *	-x	Extract files naming each as "FFFFFF.LNNN", where LNNN is the User ID and
 *			FFFFFF is the file name.
 *	-X	Extract files into directories (created as needed) as "LNNN/FFFFFF".
 *	-O	Overwrite files that already exist.
 *	-D	Write the files to stdout (or additionally, if -x or -X specified).
 *	-C	Convert the files written (via -x or -X) to plain ASCII.
 *	-S LEV	Force processing as a System Level LEV tape.  Supported LEV values are:
 *			7000	TSC-200 or TSC-300 (not yet, use 6000)
 *			6000	2000/ACCESS
 *			3500	2000/ACCESS-written 2000F image
 *			3000	2000F or 2000F'
 *			2000	2000C'
 *			1500	2000E Development In-progress
 *			1300	2000C (not yet)
 *			1200	2000B (not yet)
 *			1100	2000A (not yet)
 *		Note: Specifying an incorrect System Level can result in program crashes.
 *	-f	The input SIMH/TAP tape image disk file name.  Multiple occurences indicate
 *		multiple volumes of a tape set.
 *	-I	Input file listing filemask entries, one per line.
 *	-V	Display the program version information.
 *
 *	filemask	One or more "masks" to match filename/IDs specifying
 *		file to be acted upon.  These masks use wildcarding characters
 *		'*' (multiple-character wildcard match) and '?' (single-character
 *		wildcard match).  Generally, files are specified as FILNAM.X000.
 *		The following examples help exhibit wildcard usage:
 *			FIL*.X000	All files owned by X000 that begin with "FIL".
 *			*.X000		All files owned by X000.
 *			*.X*		All files owned by X000 thru X999.
 *			F???.X000	All files owned by X000 that have names
 *					4 characters long and begin with "F".
 *			*.X?00		All files for accounts X000, X100, X200...X900.
 *			*.* or *	All files.
 *
 *	Verbosity settings:
 *		Normally, tsb(1) produces no output.  Specifying one or more 'v' command-line
 *		flags will increase the verboseness of the programs's operations:
 *	-v	List filenames are listed as they are encountered on the DUMP portion of the tape.
 *	-vv	Produce a file listing identifical to the console "DIR" command for the system-level
 *		OS identified.
 *	-d X...	Debugging output.  Valid debugging levels (can be combined) are:
 *			H	Tape Headers/Labels
 *			S	System Processing
 *			I	Misc DeToken Conversions
 *			L	Tape/System Tables
 *			R	Raw file stats
 *			F	Files processed
 *			M	Memory allocation/deallocation
 *			P	Raw tape processing
 *			D	Detokenization
 *			C	ASCII conversions
 *			E	Finer Detokenization info
 *			T	Tokenization info
 *		The debugging output can be phenominally verbose.
 */

/*
 * BUILDING:
 *	FreeBSD, NetBSD, OpenBSD, Linux, Solaris:
 *		% gcc -o tsb *.c -lm
 *	IRIX:
 *		% gcc -o tsb *.c -lgen -lm
 *	Cray:
 *		% module load PrgEnv
 *		% cc -o tsb *.c
 *
 * PLATFORMS:
 * 
 *	Built and tested on:
 *
 *		FreeBSD 4.x, 5.x and 6.x, Intel i386 and amd64 64-bit..
 *		NetBSD 1.4, 1.5, 1.6 and 2.0, i386, amd64, sparc, sparc64, sgimips,
 *			 hp300 and vax tested.  All other platforms should also work.
 *		Mac OS X 10.3, PowerPC.
 *		SuSE Linux SLES 8 and 9, Intel i386 and amd64 64-bit.
 *		Debian Linux 2.4 kernel, Intel i386.
 *		SGI IRIX 6.5, MIPS 32-bit and MIPS 64-bit.
 *		Cray UNICOS, Cray C90/T90/J90/SV1/Y-MP PVP 64-bit Vector platforms.
 *		Cray UNICOS/mp, Cray X1/X1E Vector 64-bit.
 *		Cray UNICOS/lc, Cray XT3 64-bit.
 *		Cray UNICOS/mk, Cray T3E 64-bit.
 *		Cray HPC enhanced Linux on Cray XD1 (Linux), Cray XD1 64-bit.
 *		Sun Solaris 2.6 thru 2.10, Sparc 32-bit and 64-bit.
 *
 *	Unported:
 *
 *		OpenVMS
 *		Microsoft
 */

/*
 * SUPPORTED SYSTEM TAPES:
 *
 *	SYSTEM LEVEL		POSSIBLE TAPES	SUPPORT
 *
 *	TSC-300			DUMP/HIB/SLEEP	Full support.  (same as HP-2000F)
 *	TSC-200			DUMP/HIB/SLEEP	Full support.  (same as HP-2000F)
 *	2000/ACCESS 1A & IB	DUMP/HIB/SLEEP	Full support.
 *	2000/F & 2000/F-200	DUMP/HIB/SLEEP	Full support.
 *	2000/E			DUMP		Full support.
 *	2000/C'			DUMP/HIB/SLEEP	Full support, untested.
 *	2000/C			DUMP		* Not available.
 *	2000/B			DUMP		* Not available.
 *	2000/A			DUMP		* Not available.
 *
 *		(*) Not available.  Will be supported once sample dumps are obtained.
 *
 * FUTURE SUPPORT:
 *
 *	o	Extraction of files from raw disk images, all systems.
 *	o	Re-tokenization of ASCII files.
 *	o	Generation of 2000F- and ACCESS-image tapes from local
 *		disk files to ease migration of programs into real or
 *		simulated systems.  (See simh.trailing-edge.com for a
 *		working HP-21XX emulator.)
 */

/*
 *	If you need technical assistance with this program, have any
 *	HP-2000A, HP-2000B, HP-2000C or HP-2000E tapes or images you
 *	can provide for testing, or just want to comment on this program,
 *	please contact Scott G. Taylor at staylor@mrynet.com.
 */

/*
 * NOTES:
 *   System level
 *				6000    ACCESS IB
 *				5000    ACCESS 1A
 *				3500	2000/F Options 002 and 003.
 *				3000    2000/F
 *				2000    2000/C'
 *   Feature level		2000
 *				1500    ACCESS/IB
 *				1000    ACCESS/1A
 *				250     2000/F
 *				200     2000/F, 2000/C'
 */

/*
 * TODO:
 * o	Implement functionality to read an already-extracted program image or
 *	file and convert to ASCII.
 * o	Implement functionality to read already-extracted program images or
 *	files and create a 2000F or 2000/ACCESS tape image suitable for loading
 *	onto a running system.
 * o	Implement functionality to allow local ASCII programs to be tokenised
 *	and written back to disk or DUMP tape.
 * o	Develop an interactive mode to allow programs and files to be copied or
 *	converted out interactively.  Provide DIR, COPY, SAVE, CSAVE and LIST
 *	commands.   This will effectively require the entire tape to be imaged
 *	in memory... lots of memory--or possibly multiple passes of the tape.
 * o	Convert 2000A/B/C/E tapes to 2000F, and 2000F tapes to ACCESS?  This
 *	can first be via ASCII files, and later via retokenization.
 * o	Create routine to convert 2-byte (word) disk address to dtype, drive,
 *	track and sector for HP-2000E.
 * o	Create routine to return a populated string of
 *		"Removable|Fixed Drive %d, Track %d, Sector %d" for HP-2000E.
 * o	Continually add error-checking to avoid program crashes when bad or
 *	unknown tapes and images are processed.  In particular, watch for
 *	impossible tokenizations, line numbers, symbol tables and statement
 *	lengths; invalid IDs and passwords; equipment table and other header
 *	values out of range; etc.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"

#include "tape.h"
#include "tsb.h"
#include "hptsb.h"
#include "decl.h"
#include "version.h"

uint32_t	last_progID = 0;

#define MAXENTRIES	512

uint32_t	nentries = 0;
uint8_t		rentry[MAXENTRIES][12];

/*
 * usage -	Display program usage.
 */

void
usage (void)
{
 	errorexit (1, "Usage: %s [-xXCDOvV] [-d x...] [-S LEV] -f filename [-I inputlistfile | filemask ...]\n",
		progname);
}

int
main (int argc, char **argv)
{
	char		buf[BUFMAX];
	char		listfilename[MAXPATHLEN] = "";
	int		recno = 0;
	int		fileno = 0;
	uint32_t	rlen;
	int		c;
	int		EOFcount = 0;		/* Number of consecutive EOFs read */
	int		read_state = 0;		/* Processing thread state */
	int		i, j;
	int		Foff;			/* offset to skip over reclen in 2000F recs */
	struct tapeinfo	tio;
	uint32_t	sysEtrks = 0;
	uint32_t	Etrki = 0;
	uint8_t 	Eentry[47 * 256];	/* Buffer to hold the largest possible 2000E file/program */
	struct Etrklist {
		uint32_t	track;
		uint32_t	count;
		uint32_t	pos;
	} Etrklist[202];

	progname = basename (argv[0]);

	dprintf (DEBUG_FILEREAD, "sbrk(0) = %d\n", sbrk(0));
	while ((c = getopt(argc, argv, "d:CDxXS:Of:vI:V")) != -1) {
		switch (c) {
		case 'X':
			extract++;
			writedir++;
		case 'x':
			extract++;
			break;
		case 'O':
			overwrite++;
			break;
		case 'f':
			fncount++;
			strcpy (infilename[fncount], optarg);
			break;
		case 'v':
			verbose++;
			break;
		case 'd':
		    {
			struct debuglist	*dl;

			for (i = 0; i < strlen(optarg); i++) {
				dl = debuglist;
				while (dl->level != 0) {
					if (optarg[i] == dl->letter) {
						debug += (2 << dl->level);
						break;
					}
					dl++;
				}
				if (dl->level == 0) {
					eprintf ("Invalid debug level: %c\n", optarg[i]);
					exit(99);
				}
			}
			break;
		    }
		case 'C':
			convert++;
			break;
		case 'D':
			display++;
			break;
		case 'S':
			forcelevel=atoi(optarg);
			break;
		case 'I':
			strcpy(listfilename, optarg);
			break;
		case 'V':
			eprintf ("%s\n", version);
			break;
		default:
			errorexit (1, "Invalid option: -%c\n", c);
		}
	}

	argc -= optind;
	argv += optind;

	if (listfilename[0] != 0) {
		uint8_t	*tbp;
		FILE	* flfd;
		flfd = fopen (listfilename, "r");
		if (flfd == NULL)
			errorexit (errno, "Error opening list file %s: %s\n",
					listfilename, strerror(errno));
		while ((tbp = fgets(rentry[nentries], 12, flfd)) != NULL) {
			dprintf (DEBUG_TABLES, "Adding list entry #%d: %s\n", nentries, rentry[nentries]);
			for (i = 0; i < strlen(rentry[nentries]); i++) {
				if (rentry[nentries][i] == '\n')
					rentry[nentries][i] = 0;
				else
                        		rentry[nentries][i] = toupper(rentry[nentries][i]);
			}
			nentries++;
		}
	} else if (argc == 0) {
		nentries = 1;
		strcpy(rentry[0], "*");
	} else while (argc) {
		if (strlen(argv[0]) > 11)
			errorexit (7, "Requested entry '%s' too long.   Format is FILNAM or FILNAM.ACCT\n",
					argv[0]);
		dprintf (DEBUG_SYSTEM, "Saving requested entry %s in slot %d\n", argv[0], nentries);
		strcpy (rentry[nentries], argv[0]);
		for (i = 0; i < strlen(rentry[nentries]); i++)
			rentry[nentries][i] = toupper(rentry[nentries][i]);
		nentries++;
		argc -= 1;
		argv++;
	}

	if (fncount == -1) {
		eprintf ("No input file specified.\n");
		usage ();
	}

	if (argc != 0)
		usage();

	if (forcelevel && ! MACH_TYPE_VALID(forcelevel))
		errorexit (-1, "Invalid System Level %d.\n");

	fnno = 0;
	read_state = -1;

	tio.maxbsize = MAX_TAPE_RECLEN;
	tio.debug = (debug & (2 << DEBUG_TAPE)) != 0;


	while (1) {
		if (read_state < 0) {
			tio.fd = open (infilename[fnno], O_RDONLY);
			if (tio.fd < 0)
				errorexit (errno, "Error opening input file %s: %s\n",
						infilename[fnno], strerror(errno));

			dprintf (DEBUG_HEADERS, "MAIN: Reading tape volume %d: %s\n", fnno + 1, infilename[fnno]);
			if (read_state == -1)
				read_state = 0;
			else if (read_state == -2)
				read_state = 511;
			else
				read_state = 512;
		}
		
		if (read_state == 9999)
			break;
		dprintf (DEBUG_FILEREAD, "*** Read state is: %d\n", read_state);

		/*
		 * Read the next tape record.
		 */

		rlen = tape_read (&tio, buf, tio.maxbsize);

		if (rlen < 0) {
			dprintf (DEBUG_SYSTEM, "MAIN: Physical end of volume encountered.\n");
			if (read_state == 51) {
				eprintf ("MAIN: Unexpected physical end of volume encountered.  Can't proceed.\n");
				break;
			}
		}

		if (rlen == 0) {
			dprintf (DEBUG_SYSTEM, "MAIN: EOF mark encountered.\n");
			EOFcount++;
			if (EOFcount > 1) {
				dprintf (DEBUG_SYSTEM, "MAIN: Logical EOT detected.\n");
				break;
			}
			if (read_state != 51)
				recno = 0;
		} else {
			if (recno == 0) {
				fileno++;
				dprintf (DEBUG_SYSTEM, "MAIN: File #%d\n", fileno);
			}
			recno++;
			dprintf (DEBUG_SYSTEM, "File %d rec %d: Read %d bytes\n", fileno, recno, rlen);
			EOFcount = 0;
		}

		if (MACH_TYPE_2000E(tape_mach_type) && rlen && (65536 - WTOI(buf)) * 2 + 2 != rlen)
			dprintf (DEBUG_SYSTEM, "WARNING: Corrupt record.  Embedded reclen of %d (0x%04x) does not match record length\n",
					(65536 - WTOI(buf)) * 2 + 2, WTOI(buf));

		if (rlen > 0 && verbose > 8) {
			eprintf ("Record data:\n");
			hexdump (buf, rlen);
		}

		/*
		 * Process the tape record.
		 */

		if (rlen == 0 && read_state < 32 && read_state != 1) {
			dprintf (DEBUG_SYSTEM, "MAIN: Unexpected EOF in SYSTEM file.\n");
			read_state = 48;
		}
		
		switch (read_state) {
		case 0:
			if (rlen == 12 && !strncmp(buf + 4, "TSSYSTEM", 8)) {
				/* 2000E/2000C/2000B/2000A tape label */
				sysEtrks = 0;
				tape_mach_type = MACH_TYPE_SET_2000E;
				tape_system_level = MACH_TYPE_SET_2000E;
				if (forcelevel) {
					dprintf (DEBUG_HEADERS, "MAIN: forcing System Level processing from %d to %d\n",
							tape_mach_type, forcelevel);
					tape_system_level = forcelevel;
					tape_mach_type = MACH_TYPE_SET_2000E;
				}
				tape_reelno = 0;
				tape_date_year = 0;
				tape_date_hour = 0;
				tape_date_tsec = 0;
				tape_system_level = WTOI (buf + 2); /* Actually the System ID */
				tape_feature_level = 0;
				read_state = 1;
				dprintf (DEBUG_HEADERS, "MAIN: HP-2000E SLEEP tape.  System ID: %d\n",
						tape_system_level);
				IDtablesize = 0;
			} else if (rlen == 20 && !strncmp(buf + 2, "LBTS", 4)) {
				/* 2000C'/2000F/2000ACCESS/2000TSC tapes */
				tape_system_level = WTOI (buf + 16);
				if (forcelevel) {
					dprintf (DEBUG_HEADERS, "MAIN: forcing System Level processing from %d to %d\n",
							tape_system_level, forcelevel);
					tape_system_level = forcelevel;
				}
				switch (tape_system_level) {
#ifdef notyet
				case 12549:	tape_mach_type = MACH_TYPE_SET_TSC;
						break;
#endif
				case 6000:	tape_mach_type = MACH_TYPE_SET_ACCESSB;
						break;
				case 12549:
				case 5000:	tape_mach_type = MACH_TYPE_SET_ACCESSA;
						break;
				case 3500:	tape_mach_type = MACH_TYPE_SET_2000FP;
						break;
				case 3000:	tape_mach_type = MACH_TYPE_SET_2000F;
						break;
				default:	tape_mach_type = MACH_TYPE_SET_UNKNOWN;
						dprintf (DEBUG_HEADERS, "MAIN: Undefined 2000F+ tape type: %d.\n",
								tape_system_level);
						break;
				}
				tape_reelno = WTOI(buf + 8);
				tape_date_year = WTOI(buf + 10);
				tape_date_hour = WTOI(buf + 12);
				tape_date_tsec = WTOI(buf + 14);
				tape_system_level = WTOI(buf + 16);
				tape_feature_level = WTOI(buf + 18);
				read_state = 1;
				dprintf (DEBUG_HEADERS, "MAIN: Tape System Level: %s\n",
						machtypename(tape_system_level));
				dprintf (DEBUG_HEADERS, "    Reel No:   %d\n", tape_reelno);
				dprintf (DEBUG_HEADERS, "       Date:   ");
				if (verbose > 2)
					day2date (tape_date_year, tape_date_hour, tape_date_tsec);
				dprintf (DEBUG_HEADERS, " [Y%d/H%d/T%d]\n",
						tape_date_year, tape_date_hour, tape_date_tsec);
				dprintf (DEBUG_HEADERS, " System Lev:   %d\n", tape_system_level);
				dprintf (DEBUG_HEADERS, "Feature Lev:   %d\n", tape_feature_level);
			} else {
				dprintf (DEBUG_SYSTEM, "MAIN: Unrecognized tape label.\n");
			}
			if (tape_reelno > 1)
				read_state = 50;
			break;
		
		case 1: if (rlen == 0) {
				dprintf (DEBUG_HEADERS, "MAIN: DUMP format tape.\n");
				read_state = 48;
			} else {
				dprintf (DEBUG_HEADERS, "MAIN: SLEEP or HIBERNATE format tape.\n");
				read_state = 5;
				if (MACH_TYPE_ACCESS(tape_mach_type))
					eval_2000access_equipment_table (buf, rlen);
				else if (MACH_TYPE_2000F(tape_mach_type))
					eval_2000F_equipment_table (buf, rlen);
				else if (MACH_TYPE_2000C(tape_mach_type))
					eval_2000C_equipment_table (buf, rlen);
				else if (MACH_TYPE_2000E(tape_mach_type)) {
					eval_2000E_equipment_table (buf, rlen);
					read_state = 200;
				} else {
					dprintf (DEBUG_SYSTEM, "MAIN: Can't handle equipment table for unknown machine type.  Skipping system records.\n");
					read_state = 48;
				}
			}
			break;

		case 5: if (MACH_TYPE_ACCESS(tape_mach_type)) {
				if (rlen != 1200) {
					dprintf (DEBUG_SYSTEM, "MAIN: Expected 1200 bytes for 2000/ACCESS-format DIREC table.  Read %d bytes.\n", rlen);
					read_state = 48; /* State 48 means "skip all records until EOF, then assume DUMP data */
				} else {
					eval_ACCESS_DIREC_table (buf, rlen);
					read_state = 6;
				}
			} else if (MACH_TYPE_2000F(tape_mach_type)) {
				if (rlen != 1122) {
					dprintf (DEBUG_SYSTEM, "MAIN: Expected 1122 bytes for 2000F-format DIREC table.  Read %d bytes.\n", rlen);
					read_state = 48; /* State 48 means "skip all records until EOF, then assume DUMP data */
				} else {
					eval_2000C_DIREC_table (buf, rlen);
					read_state = 7;
				}
			}
			break;
		case 6: if (MACH_TYPE_ACCESS(tape_mach_type)) {
				if (rlen != 642) {
					dprintf (DEBUG_SYSTEM, "MAIN: Expected 642 bytes for 2000/ACCESS-format non-shareable device table.  Read %d bytes.\n", rlen);
					read_state = 48;
				} else {
					dprintf (DEBUG_TABLES, "MAIN: NON-SHARABLE DEVICE TABLE record read as expected--BUT NOT REALLY!!!.\n");
					eval_2000access_NONSHAREDEVICE_table (buf, rlen);
#ifdef NOTYET
					display_2000access_NONSHAREDEVICE_table ();
#endif
					read_state = 7;
				}
			} else if (MACH_TYPE_2000F(tape_mach_type)) {
				dprintf (DEBUG_SYSTEM, "MAIN: Can't handle 2000F system records... Skipping to files.\n");
				read_state = 7;
			}
			break;
		case 7: if (rlen == 2048) {
				dprintf (DEBUG_FILEREAD, "MAIN: ID table record read.  Expecting more.\n");
				load_ID_table (buf, rlen, tape_mach_type, &IDtablesize);
				dprintf (DEBUG_FILEREAD, "MAIN: NEW ID table size = %d\n", IDtablesize);
			} else if (rlen > 2048) {
				dprintf (DEBUG_FILEREAD, "MAIN: ID table record too big: size=%d\n", rlen);
				read_state = 48;
			} else {
				dprintf (DEBUG_FILEREAD, "MAIN: ID table record read.  No more expected.\n");
				load_ID_table (buf, rlen, tape_mach_type, &IDtablesize);
				dprintf (DEBUG_FILEREAD, "MAIN: ID table completed.  ID table listing:\n");
				if (verbose > 2) 
					dump_ID_table (IDtable, IDtablesize, tape_mach_type);
				read_state = 8;
			}
			break;
		case 8: 
			if (Dtab_total == 0) {
				DIRtab = calloc (1, sizeof (struct BlockTable));
				DIRtab->len = 0;
				dprintf (DEBUG_HEADERS, "MAIN: Directory table.\n");
			}

			Foff = 0;
			if (MACH_TYPE_2000F(tape_mach_type))
				Foff = 2;

			if (DIRtab->len == 0) {
				dprintf (DEBUG_TABLES, "MAIN: Initializing Dtab.\n");
				Dtab = DIRtab;
				Dtab_total = 0;
			} else {
				dprintf (DEBUG_TABLES, "MAIN: Adding entry to Dtab.\n");
				Dtab->next = calloc (1, sizeof (struct BlockTable));
				Dtab = Dtab->next;
			}
			dprintf (DEBUG_TABLES, "MAIN: Incrementing Dtab_total of %d by %d\n",
				Dtab_total, rlen - Foff);
			Dtab->len = rlen - Foff;
			Dtab_total = Dtab_total + Dtab->len;
			Dtab->block = calloc (Dtab->len, sizeof (uint8_t));
			dprintf (DEBUG_TABLES, "MAIN: Copying %d bytes into Dtab entry.\n", Dtab->len);
			memcpy (Dtab->block, buf + Foff, Dtab->len);
			Dtab->next = NULL;

			if (Dtab->len >= 2048)
				break;
			Dtab = DIRtab;
			dprintf (DEBUG_TABLES, "MAIN: allocating new Dtab block of %d bytes.\n",
					Dtab_total);
			nDtab = calloc(Dtab_total, sizeof (uint8_t));
			dprintf (DEBUG_TABLES, "      New Dtab address: %x\n", nDtab);
			nDtabp = nDtab;

			while (Dtab != NULL) {
				dprintf (DEBUG_TABLES, "MAIN: Copying over %d bytes to new Dtab at %x\n",
						Dtab->len, nDtabp);
				memcpy (nDtabp, Dtab->block, Dtab->len);
				nDtabp = nDtabp + (Dtab->len);
				nfree (Dtab->block);
				Dtabp = Dtab->next;
				nfree (Dtab);
				Dtab = Dtabp;
				dprintf (DEBUG_TABLES, "MAIN: New nDtab is %x, len: %d\n",
						nDtabp, nDtabp - nDtab);
			};

			dprintf (DEBUG_TABLES, "MAIN: Total of %d bytes copied into new Dtab\n",
					nDtabp - nDtab);

			if (verbose > 2) {
				int i, iascii,ifile,icomp,i4,i5;
				for (i = 0; i < (nDtabp - nDtab) / 24; i++) {
					if (WTOI(nDtab + i * 24 + 000 * 2) == 0 ||
						    WTOI(nDtab + i * 24 + 000 * 2) == 0xffff)
						continue;
					iascii = ifile = icomp = i4 = i5 = 0;
					if (*(nDtab + i * 24 + 001 * 2) & 0x80) iascii++;
					*(nDtab + i * 24 + 001 * 2) &= 0x7f;
					if (*(nDtab + i * 24 + 1 + 001 * 2) & 0x80) icomp++;
					*(nDtab + i * 24 + 1 + 001 * 2) &= 0x7f;
					if (*(nDtab + i * 24 + 2 + 001 * 2) & 0x80) ifile++;
					*(nDtab + i * 24 + 2 + 001 * 2) &= 0x7f;
					if (*(nDtab + i * 24 + 3 + 001 * 2) & 0x80) i4++;
					*(nDtab + i * 24 + 3 + 001 * 2) &= 0x7f;
					if (*(nDtab + i * 24 + 4 + 001 * 2) & 0x80) i5++;
					*(nDtab + i * 24 + 4 + 001 * 2) &= 0x7f;
					eprintf ("ID: %c%03d N:%6.6s F:%d%d%d%d%d S:%06o Y:%d R:%.4d s:%.5d DA:%.5d D:%.5d A:%.5d l:%.5d\n",
						WTOI(nDtab + i * 24 + 000 * 2) / 1024 + '@',
						WTOI(nDtab + i * 24 + 000 * 2) % 1024,	/* ID */
						nDtab + i * 24 + 001 * 2,		/* Filnam*/
						iascii, ifile, icomp, i4, i5,		/* Flags */
						WTOI(nDtab + i * 24 + 4 * 2),		/* Start addr/Recs */
						WTOI(nDtab + i * 24 + 5 * 2),		/* Create Date */
						WTOI(nDtab + i * 24 + 6 * 2),		/* Mod Date */
						WTOI(nDtab + i * 24 + 7 * 2),		/* Status */
						WTOI(nDtab + i * 24 + 8 * 2),		/* Disk Addr */
						WTOI(nDtab + i * 24 + 9 * 2),		/* Devno */
						WTOI(nDtab + i * 24 + 10 * 2),		/* Loader info */
						WTOI(nDtab + i * 24 + 11 * 2));		/* Len */
				}
			}
			read_state = 9;
			nfree(nDtab);
			/*nfree(DIRtab);*/
			break;
		case 9:
			/* Process 2000F SYSTEM SEGMENT TABLE */
			/*
			 * The System Segment Table (SST) is a 53-word table resident in
			 * the loader.  It is the first portion of the bootstrap and is
			 * pointed to by LDBSA.  The first word of the table contains the
			 * negative number of system segments.  Each group of 4 words
			 * following the first word has the following format:
			 *  WORD 1	length of segment in -words
			 *       2	absolute, beginning core address of the segment
			 *	3-4	disc address of the segment
			 *
			 * There are 13 segments, ordered as follows:  (8 seen)
			 *  SEG	1	Interrupt Locations -- 2 to 27(8)
			 *	2	System base page -- end of EQT to 1777(8)
			 *	3	System linkage area -- 2002 to 2015(8)
			 *	4	System segment 1 -- End of DIREC to 41777(8)
			 *	5	  "      "     2 -- 42000 to 51777(8)
			 *	6	  "      "     3 -- 52000 to 61777(8)
			 *	7	  "      "     4 -- 62000 to 71777(8)
			 *	8	  "      "     5 -- 72000 to 77677(8)
			 *	9	Equipment Table -- 100 to 216(8)
			 *	10	DIREC table -- 30000 to 31057(8)
			 *	11	Loader Segment 2 -- 16000 to 25777(8)
			 *	12	Laoder Segment 2 - 2000 to 14677(8)
			 *	13	Disc driver -- 26000 to 27777(8)
			 *
			 * Note that this includes all core resident portions of the
			 * loader and system except for locations 14700 to 15777(8).
			 * The first 1000(8) of these words comprise the disc bootstrap
			 * and are resident on blocks 1 and 2 of each disc.  Locations
			 * 15700 to 15777(8) may only be used for tempraries.
			 */

			Foff = 0;
			if (MACH_TYPE_2000F(tape_mach_type))
				Foff = 2;

			SST.count = 65536 - WTOI (buf + Foff);
			dprintf (DEBUG_HEADERS, "MAIN: System Segment Table (SST).  %d entries\n",
					SST.count);
			for (i = 0; i < SST.count; i ++) {
				SST.rec[i].len = 65536 - WTOI (buf + Foff + 2 + i * 8 + 0);
				SST.rec[i].start = WTOI (buf + Foff + 2 + i * 8 + 2);
				SST.rec[i].addr[0] = WTOI (buf + Foff + 2 + i * 8 + 4);
				SST.rec[i].addr[1] = WTOI (buf + Foff + 2 + i * 8 + 6);
				dprintf (DEBUG_HEADERS, "    %d:\tlen=%05d (%z6d) \tMem: %06o-%06o\tDiscAddr=%06o/%06o\n",
						i,
						SST.rec[i].len,
						SST.rec[i].len * 2,
						SST.rec[i].start,
						SST.rec[i].start + SST.rec[i].len - 1,
						SST.rec[i].addr[0], SST.rec[i].addr[1]);

			}
			Srecno = 0;
			read_state = 10;
			break;
		case 10:
			Foff = 0;
			if (MACH_TYPE_2000F(tape_mach_type))
				Foff = 2;

			Srecno++;
			if (rlen - Foff != SST.rec[Srecno - 1].len * 2) {
				eprintf ("MAIN: Expecting SST segment record #%d of %d bytes -- got %d\n",
						Srecno, SST.rec[Srecno - 1].len * 2, rlen - Foff);
				eprintf ("MAIN: Skipping the rest of the system records.\n");
				read_state = 48;
				break;
			}
			SST.rec[Srecno - 1].block =
					calloc(SST.rec[Srecno - 1].len * 2, sizeof (uint8_t));
			memcpy(SST.rec[Srecno - 1].block, buf + Foff, SST.rec[Srecno - 1].len * 2);
			dprintf (DEBUG_TABLES, "MAIN: GOT SST Segment record #%d as expected.\n",
					Srecno);
#ifdef COMMAND_TABLE_HACK
			for (i = 0; i < SST.rec[Srecno - 1].len - 2; i++)
				eprintf ("--------%d  %c%c%c\n",
					i,
					((WTOI(SST.rec[Srecno - 1].block + i * 2) >> 10) & 037) + 0101,
					((WTOI(SST.rec[Srecno - 1].block + i * 2) >> 5) & 037) + 0101,
					(WTOI(SST.rec[Srecno - 1].block + i * 2) & 037) + 0101);
#endif
			if (Srecno == SST.count)
				read_state = 11;
			break;
		case 11:
			/* Got all the SST records wanted.   Release the memory now. */
			for (i = 0; i < Srecno; i++)
				nfree (SST.rec[i].block);
			Foff = 0;
			if (MACH_TYPE_2000F(tape_mach_type))
				Foff = 2;

			SLT.count = 65536 - WTOI (buf + Foff) - 1;
			dprintf (DEBUG_HEADERS, "MAIN: System Library Table (SLT).  %d entries\n",
					SLT.count);
			for (i = 0; i < SLT.count; i ++) {
				SLT.rec[i].len = 65536 - WTOI (buf + Foff + i * 2 + 2);
				SLT.rec[i].start = 0;
				SLT.rec[i].addr[0] = 0;
				SLT.rec[i].addr[1] = 0;
				dprintf (DEBUG_HEADERS, "    %d:\tlen=%05d (%04x)\tstart=%06o\tAddr=%06o/%06o (%d)\n",
						i,
						SLT.rec[i].len,
						65536 - SLT.rec[i].len,
						SLT.rec[i].start,
						SLT.rec[i].addr[0], SLT.rec[i].addr[1],
						SLT.rec[i].len * 2);

			}
			Srecno = 0;
			read_state = 12;
			break;
		case 12:
			Foff = 0;
			if (MACH_TYPE_2000F(tape_mach_type))
				Foff = 2;

			Srecno++;
			if ((MACH_TYPE_2000F(tape_mach_type) || MACH_TYPE_2000C(tape_mach_type))
					&& rlen - Foff != SLT.rec[Srecno - 1].len * 2) {
				eprintf ("MAIN: Expecting SLT segment record #%d of %d bytes -- got %d\n",
						Srecno, SLT.rec[Srecno - 1].len * 2, rlen - Foff);
				eprintf ("MAIN: Skipping the rest of the system records.\n");
				read_state = 48;
				break;
			}
			if (MACH_TYPE_ACCESS(tape_mach_type)
					&& rlen < SLT.rec[Srecno - 1].len * 2) {
				eprintf ("MAIN: Expecting SLT segment record #%d of at least %d bytes -- got %d\n",
						Srecno, SLT.rec[Srecno - 1].len * 2, rlen - Foff);
				eprintf ("MAIN: Skipping the rest of the system records.\n");
				read_state = 48;
				break;
			}
			SLT.rec[Srecno - 1].block = 
					calloc(SLT.rec[Srecno - 1].len, sizeof (uint8_t));
			memcpy(SLT.rec[Srecno - 1].block, buf + Foff, SLT.rec[Srecno - 1].len);
			dprintf (DEBUG_TABLES, "MAIN: GOT SLT Segment record #%d as expected.\n",
					Srecno);
			if (Srecno == SLT.count)
				read_state = 48;
			break;
		case 200:	/* 2000E IDT tracks */
			dprintf (DEBUG_TABLES, "MAIN: 2000E IDT tracks record\n");
			for (i = 0; i < (rlen - 2) / 8; i++)
				if (WTOI(buf + (i * 8) * 2 + 2))
					dprintf (DEBUG_TABLES, "    %d: First ID: %c%03d  Password: %6.6s\n",
							i,
							WTOI(buf + (i * 8) * 2 + 2) / 1024 + '@',
							WTOI(buf + (i * 8) * 2 + 2) & 0x3ff,
							buf + (i * 8 + 1) * 2 + 2);
			read_state++;
			sysEtrks = 0;
			Etrki = 0;
			break;
		case 201:	/* 2000E System Disc Track Length Table */
		case 208:	/* 2000E User Disc Track Length Table */
			for (i = 0; i < 203; i++)
				for (j = 0; j < 48; j++)
					Etrack[i][j] = NULL;

			dprintf (DEBUG_TABLES, "MAIN: 2000E Track Length Table record\n");
			for (i = 0; i < (rlen - 2) / 2; i++) {
				if (WTOI (buf + (i * 2) + 2)) {
					dprintf (DEBUG_TABLES, "    Track %d:  %d Sectors/%d Bytes/%d Tape Records\n",
							i,
							WTOI (buf + (i * 2) + 2),
							WTOI (buf + (i * 2) + 2) * 256,
							WTOI (buf + (i * 2) + 2) * 256 / 6144 +
								(WTOI (buf + (i * 2) + 2) * 256 % 6144 != 0));
					sysEtrks +=  WTOI (buf + (i * 2) + 2) * 256 / 6144 +
							(WTOI (buf + (i * 2) + 2) * 256 % 6144 != 0);
					Etrklist[Etrki].track = i;
					Etrklist[Etrki].count = WTOI (buf + (i * 2) + 2);
					Etrklist[Etrki].pos = 0;
					Etrki++;
					/* Allocate this track */
					for (j = 0; j < WTOI (buf + (i * 2) + 2); j++)
						Etrack[i][j] = calloc(256, sizeof(uint8_t));
				}
			}
			read_state++;
			Etrki = 0;
			break;
		case 202:	/* 2000E A000/User Library Tracks */
		case 209:	/* 2000E User Disc User Library Tracks */
			if (sysEtrks) {
				dprintf (DEBUG_TABLES, "MAIN: 2000E A000/User Library Tracks\n");
				sysEtrks--;
				for (i = 0; i < (rlen - 2) / 256; i++) {
					dprintf (DEBUG_TABLES, "MAIN: Copying 256 bytes from pos %d to Track %d, sector %d\n",
							i * 256,
							Etrklist[Etrki].track,
							Etrklist[Etrki].pos + i);
					memcpy(Etrack[Etrklist[Etrki].track][Etrklist[Etrki].pos + i],
							buf + 2 + i * 256, 256);
				}
				Etrklist[Etrki].pos += (rlen - 2) / 256;
				dprintf (DEBUG_TABLES, "MAIN: Copied %d sectors, Sector count now %d\n",
						(rlen - 2) / 256, Etrklist[Etrki].pos);
				if (Etrklist[Etrki].pos == Etrklist[Etrki].count)
					Etrki++;
				if (Etrklist[Etrki].pos > Etrklist[Etrki].count)
					eprintf ("MAIN: *** Read track too large -- Expect a crash.\n");
				dprintf (DEBUG_TABLES, "MAIN: 2000E A000/User Library Tracks left: %d\n",
						sysEtrks);
				break;
			}
			read_state++;
		case 203:	/* 2000E System Disc Directory Tracks */
		case 210:	/* 2000E User Disc Directory Tracks */
			if (read_state == 203 || verbose > 2) {
				if (read_state == 203)
					dprintf (DEBUG_TABLES, "MAIN: 2000E A000/User Directory Tracks\n");
				else
					dprintf (DEBUG_TABLES, "MAIN: 2000E User Disc Directory Tracks\n");
				if (verbose > 1 || debug) {
					eprintf (" ID    NAME    DATE  SUB  TR/SEC   LEN   FP");
					dprintf (DEBUG_HEADERS, "     START  FIXED");
					eprintf ("\n");
				}
			}
			last_progID = 0;
			for (i = 0; i < (rlen - 2) / 16; i++) {
				uint8_t	entry[12];

				uint32_t	trk, sect, dtype, drive;
				uint32_t	prot = 0, file = 0;
				uint32_t	id;
				uint8_t		ID[5];

				if (!WTOI(buf + (i * 8) * 2 + 2) ||
						WTOI(buf + (i * 8) * 2 + 2) == 0xffff)
					continue;

				dtype = (WTOI(buf + (i * 8 + 6) * 2 + 2) >> 6) & 0x01;
				drive = WTOI(buf + (i * 8 + 6) * 2 + 2) >> 15;
				trk = (WTOI(buf + (i * 8 + 6) * 2 + 2) >> 7) & 0xff;
				sect = WTOI(buf + (i * 8 + 6) * 2 + 2) & 0x3f;
				Pentry.len = 65536 - WTOI(buf + (i * 8 + 7) * 2 + 2);
				Pentry.filesize = Pentry.len * 2;
				Pentry.symtabstart = Pentry.len;
				Pentry.block_size = Pentry.filesize;
				Pentry.block = Eentry;
				Pentry.id = WTOI(buf + (i * 8) * 2 + 2);

				memcpy (Pentry.HPfilnam, buf + (i * 8 + 1) * 2 + 2, 6);
				memcpy (Pentry.filnam, Pentry.HPfilnam, 6);

				Pentry.filnam[6] = ' ';
				*(index(Pentry.filnam, ' ')) = 0;

				if (Pentry.filnam[0] & 0x80) {
					prot = 1;
					Pentry.filnam[0] &= 0x7f;
				}
				if (Pentry.filnam[2] & 0x80) {
					file = 1;
					Pentry.filnam[2] &= 0x7f;
				}
				if (file) {
					Pentry.type = ENTRY_TYPE_SET_FILE;
					Pentry.startloc = 256;
				} else {
					Pentry.type = ENTRY_TYPE_SET_PROGRAM;
					Pentry.startloc = WTOI(buf + (i * 8 + 4) * 2 + 2);
				}

				/* Fill out a header if needed for an HP-image file extraction */
				memcpy (Pentry.header, buf + (i * 8) * 2 + 2, 2);		/* ID	  */
				memcpy (Pentry.header + 2, Pentry.filnam, 6);			/* FILNAM */
				Pentry.header[4] |= file ? 0x80 : 0;				/* FFLAG  */
				Pentry.header[8] = 0x04;					/* START  */
				Pentry.header[9] = 0x00;
				memcpy (Pentry.header + 10, buf + (i * 8 + 1) * 2 + 10, 2);	/* REFDATE*/
				memcpy (Pentry.header + 12, buf + (i * 8 + 1) * 2 + 12, 2);	/* MODDATE*/
				Pentry.header[14] = 0;						/* STATUS */
				Pentry.header[15] = 0;
				Pentry.header[16] = 0;						/* DADDR  */
				Pentry.header[17] = 0;
				Pentry.header[18] = 0;						/* DDEV   */
				Pentry.header[19] = 0;
				Pentry.header[20] = 0;						/* LINFO  */
				Pentry.header[21] = 0;
				memcpy (Pentry.header + 22, buf + (i * 8 + 7) * 2 + 2, 2);	/* LEN    */

				id = WTOI(buf + (i * 8) * 2 + 2);
				snprintf (ID, 5, "%c%03d",
						Pentry.id / 1024 + '@',
						Pentry.id & 0x3ff);
				sprintf (entry, "%s.%s", Pentry.filnam, ID);
				if (!check_entrylist(entry, rentry[0], 12, nentries))
					continue;

				if (verbose == 1)
					eprintf ("%s.%s\n", Pentry.filnam, ID);
				else if (verbose > 1) {
					if (last_progID != id)
						eprintf ("%s", ID);
					else
						eprintf ("    ");
					eprintf("  %-6.6s  %03d/%02d  %d   %03d/%02d    %02d   %c%c",
							Pentry.filnam,
							WTOI(buf + (i * 8 + 5) * 2 + 2) % 1024,
							WTOI(buf + (i * 8 + 5) * 2 + 2) / 1024 + 1,
							drive, trk, sect,
							(Pentry.len + 127) / 128,
							file ? 'F' : ' ',
							prot ? 'P' : ' ');
				}
				dprintf (DEBUG_HEADERS, "    %06d    %d", Pentry.startloc, dtype);
				if (verbose > 1 || debug)
					eprintf ("\n");
				last_progID = id;

				for (j = sect; j < sect + (Pentry.len * 2 + 255) / 256; j++ ) {
					dprintf (DEBUG_FILES, "MAIN:    Copying %d bytes from Track %d, Sector %d to pos %d\n",
							256, trk, j, (j - sect) * 256);
					memcpy(Eentry + (j - sect) * 256, Etrack[trk][j], 256);
				}
				dprintf (DEBUG_FILES, "MAIN: DUMP of program %6.6s, length %d:\n",
						Pentry.filnam, Pentry.filesize);
				if (verbose > 7)
					hexdump (Eentry, Pentry.filesize);
				if (display || convert) {
					if (ENTRY_TYPE_FILE(Pentry.type))
						convert_file (&Pentry);
					else
						convert_program (&Pentry);
					if (verbose > 3 || display) {
						struct line_table *lptr = &Pentry.cline;
						uint32_t	rno = 1;
						dprintf (DEBUG_FILES, "lptr = %d\n", *lptr);
						while (lptr->next != NULL) {
							dprintf (DEBUG_FILES, "T> ");
							if (ENTRY_TYPE_FILE(Pentry.type))
								dprintf (DEBUG_FILES, "Record %d:\n", rno++);
							eprintf ("%.*s\n", lptr->len, lptr->data);
							lptr = lptr->next;
						}
					}
				}
				if (extract)
					extract_entry (&Pentry);
				if (&Pentry.cline != NULL) {
					struct line_table *lptr = &Pentry.cline, *optr;
					while (lptr != NULL) {
						dprintf (DEBUG_FILEREAD, "data=%x next=%x lptr=%x\n",  
								lptr->data, lptr->next, lptr);
						if (lptr->data != NULL)
							nfree(lptr->data);
						optr = lptr;
						lptr = lptr->next;
						if (optr != &Pentry.cline) {
							if (optr != NULL)
								nfree(optr);
						}
					}
				}
				if (&Pentry.block)
					nfree (Pentry.block);
			}
			read_state++;
			break;
		case 204:	/* 2000E Post-A000/User data EOF */
		case 211:	/* 2000E Post-User Disc data EOF */
			for (i = 0; i < 203; i++)
				for (j = 0; j < 48; j++)
					if (Etrack[i][j] != NULL)
						nfree(Etrack[i][j]);

			read_state++;
			if (rlen != 0) {
				dprintf (DEBUG_SYSTEM, "MAIN: ERROR: Expecting EOF, got record of %d bytes instead.\n",
						rlen);
				/* Fall through to next state, in case the EOF is just missing */
			} else
				break;
		case 205:	/* 2000E User Disc Tape Label */
			if (strncmp (buf + 6, "TSDSK", 5)) {
				dprintf (DEBUG_SYSTEM, "MAIN: *** TSDSK User Disk Label Encountered\n");
			}
			dprintf (DEBUG_SYSTEM, "MAIN: 2000E User Disc Tape Label\n");
			read_state++;
			break;
		case 206:	/* 2000E User Disc ADT/Directory Lengths Record */
			dprintf (DEBUG_TABLES, "MAIN: 2000E User Disc ADT/Directory Lengths record\n");
			dprintf (DEBUG_TABLES, "    ADT Length: %d  Track 1 DTL: %d  Track 2 DTL: %d\n",
					65536 - WTOI(buf + 2),
					WTOI(buf + 2 + 2) ? 65536 - WTOI(buf + 2 + 2) : 0,
					WTOI(buf + 2 + 4) ? 65536 - WTOI(buf + 2 + 4) : 0);
			read_state++;
			break;
		case 207:	/* 2000E User Disc ADT Table */
			dprintf (DEBUG_TABLES, "MAIN: 2000E User Disc ADT Table\n");
			for (i = 0; i < (rlen - 2) / 4; i++)
				dprintf (DEBUG_TABLES, "        Entry %d: %s Drive %02d, Track %03d, Starting Sector %03d, Sector Count %02d\n",
						i,
						(WTOI(buf + (i * 2) * 2 + 2) >> 6) & 0x01 ?
								"Remov" : "Fixed",
						WTOI(buf + (i * 2) * 2 + 2) >> 15,
						((WTOI(buf + (i * 2) * 2 + 2) >> 7) & 0xff) - 1,
						WTOI(buf + (i * 2) * 2 + 2) & 0x3f,
						WTOI (buf + (i * 2 + 1) * 2 + 2));
			read_state++;
			sysEtrks = 0;
			Etrki = 0;
			memset (Etrklist, 0, sizeof (struct Etrklist));
			break;
		case 212:	/* 2000E System Segment Table */
			dprintf (DEBUG_TABLES, "MAIN: 2000E System Segment Table record\n");
			sysEtrks = 0;
			for (i = 0; i < 65536 - WTOI (buf + 2) - 1; i++, sysEtrks++)
				dprintf (DEBUG_TABLES, "    Segment record %d:  Length %d words (%04x), Start Address %06o\n",
						i,
						65536 - WTOI (buf + (i * 2) * 2 + 4),
						WTOI (buf + (i * 2) * 2 + 4),
						WTOI (buf + (i * 2 + 1) * 2 + 4));
			read_state++;
			break;
		case 213:	/* 2000E System Tracks */
			if (sysEtrks) {
				dprintf (DEBUG_TABLES, "MAIN: 2000E System Tracks\n");
				sysEtrks--;
				dprintf (DEBUG_TABLES, "MAIN: 2000E System Tracks left: %d\n",
						sysEtrks);
				break;
			}
			read_state++;
		case 214:	/* 2000E System Library Lengths Record */
			dprintf (DEBUG_TABLES, "MAIN: 2000E System Library Lengths Record\n");
			sysEtrks = 0;
			memset (Etrklist, 0, sizeof (struct Etrklist));
			for (i = 0; i < 65536 - WTOI (buf + 2) - 1; i++) {
				dprintf (DEBUG_TABLES, "    System Library record %02d:  Length %03d words (%04x), Start Address %06o (%04x)\n",
						i,
						65536 - WTOI (buf + (i * 2) + 4),
						WTOI (buf + (i * 2) + 4),
						WTOI (buf + (i * 2) + 2 + (65536 - WTOI(buf + 2)) * 2),
						WTOI (buf + (i * 2) + 2 + (65536 - WTOI(buf + 2)) * 2));
				sysEtrks++;
			}
			read_state++;
			break;
		case 215:	/* 2000E System Library Record */
			if (sysEtrks) {
				dprintf (DEBUG_TABLES, "MAIN: 2000E System Library Record\n");
				sysEtrks--;
				dprintf (DEBUG_TABLES, "MAIN: 2000E System Library records left: %d\n",
							sysEtrks);
				} else
					read_state++;
				break;
			case 216:	/* Final 2000E EOF should be reached */
				if (rlen != 0) {
					dprintf (DEBUG_TABLES, "Final 2000E EOF record not reached.   Found %d bytes\n",
						rlen);
				/* We'll loop forever now, complaining */
			}
			read_state = 999;
		case 48:
			/* Got all the SLT records wanted.   Release the memory now. */
			for (i = 0; i < Srecno; i++)
				nfree (SLT.rec[i].block);
			if (rlen == 0) {
				dprintf (DEBUG_FILEREAD, "MAIN: EOF reached, continuing processing.\n");
				read_state = 50;
				if (verbose > 1) {
					if (MACH_TYPE_ACCESS(tape_mach_type))
						eprintf (" ID      NAME        DATE            LENGTH    DEV/ADDR   RECORD\n");
					else if (MACH_TYPE_2000F(tape_mach_type) ||
							MACH_TYPE_2000C(tape_mach_type))
						eprintf (" ID      NAME        DATE           LENGTH      DISC      DRUM\n");
				}
			} else {
				dprintf (DEBUG_FILEREAD, "MAIN: Ignoring record %d of %d bytes.\n",
							recno, rlen);
			}
			break;
		case 49:
			read_state = 50;
			program_entry_state = 0;
			break;
		case 50:
			if (rlen == 0)
				break;
			if (rlen == 20 && !strncmp(buf + 2, "LBTS", 4)) {
				dprintf (DEBUG_HEADERS, "MAIN: End of tape volume %d; another volume expected.\n", tape_reelno);
				fnno++;
				close (tio.fd);
				if (fnno > fncount) {
					eprintf ("MAIN: No more tape volumes.  Incomplete DUMP set.\n");
					read_state = 9999;
				}
				read_state = -3;
				break;
			}
			dprintf (DEBUG_FILEREAD, "MAIN: Processing PROGRAM ENTRY record.\n");
			read_state = 51;
			load_program_entry (program_entry_state, buf, rlen, tape_mach_type);
			program_entry_state = 1;
			break;
		case 51:
			if (rlen == 0) {
				if (recno < Pentry.block_count) {
					dprintf (DEBUG_FILEREAD, "MAIN: Premature end of program entry.   Now expecting TAPE LABEL, then next tape volume.\n");
					recno--;
					read_state = 510;
					break;
				}
				dprintf (DEBUG_FILEREAD, "MAIN: PROGRAM ENTRY completed.\n");
				read_state = 50;
				load_program_entry (2, NULL, 0, tape_mach_type);
				program_entry_state = 0;
				recno = 0;
			} else {

				dprintf (DEBUG_FILEREAD, "MAIN: More PROGRAM ENTRY data for this entry.\n");
				load_program_entry (program_entry_state, buf, rlen, tape_mach_type);
			}
			break;
		case 510:
			if (rlen == 20 && !strncmp(buf + 2, "LBTS", 4)) {
				dprintf (DEBUG_HEADERS, "MAIN: End of tape volume %d; another volume expected.\n", tape_reelno);
				fnno++;
				close (tio.fd);
				if (fnno > fncount) {
					eprintf ("MAIN: No more tape volumes.  Incomplete DUMP set.\n");
					read_state = 9999;
				}
				read_state = -2;
				break;
			} else {
				eprintf ("MAIN: Was expecting a TAPE LABEL, then another tape volume.  Can't proceed.\n");
				read_state = 9999;
				break;
			}
		case 511:
		case 512:
			if (rlen == 20 && !strncmp (buf + 2, "LBTS", 4)) {
				tape_reelno++;
				if (tape_reelno != WTOI(buf + 8)) {
					eprintf ("MAIN: Was expecting tape reel number %d, got reel number %d.  Can't proceed.\n",
						tape_reelno, WTOI(buf + 8));
					read_state = 9999;
					break;
				}
#ifdef FUTURE
				/* The following really should be compared to verify the new tape is from the same set */
				tape_date_year = WTOI(buf, 10);
				tape_date_hour = WTOI(buf + 12);
				tape_date_tsec = WTOI(buf + 14);
				tape_system_level = WTOI(buf + 16);
				tape_feature_level = WTOI(buf + 18);
#endif /* FUTURE */
				dprintf (DEBUG_SYSTEM, "MAIN: Proper continuation tape encountered.  Continuing processing.\n");
				recno--;
				if (read_state == 511)
					read_state = 51;
				else
					read_state = 50;
			} else {
				eprintf ("MAIN: New TAPE HEADER label expected, but not found.   Can't proceed.\n");
				read_state = 9999;
			}
			break;
		default:
			dprintf (DEBUG_SYSTEM, "MAIN: Unexpected data record at file %d record %d of %d bytes\n", fileno, recno, rlen);
		}
	}
	close (tio.fd);
	n_list();
	exit (0);
}
